﻿
// VideoMotion Class Version 1.5

// By Dan Zen - http://www.danzen.com
// Sheridan Interactive Multimedia - http://imm.sheridanc.on.ca
// One Year Post Grad Program in Oakville Ontario
// Posted at http://imm.sheridanc.on.ca/videomotion

import flash.display.BitmapData;
import flash.geom.*;
import flash.filters.ColorMatrixFilter;
import flash.filters.BlurFilter;
import mx.transitions.Tween;
import mx.transitions.easing.None;	

// Class to handle video motion for example from a Web cam
// methods are makeCursor and makeButton
// makeCursor creates a motion cursor that follows motion
// makeButtons tracks motion cursor events on a button clip

// This could be a package of two classes to avoid the static 
// tracking of cursor and button "instances"
// but could not seem to make the constructor return a movieclip
// so used a method instead and decided to be consistent with the button

class VideoMotion {
	
	// ------------------ motion cursor vars --------------------
	
	public var left:Number; // position of apparent the camera window
	public var top:Number;
	public var right:Number;
	public var bottom:Number;
		
	private static var cursorNum:Number = 0; // static var for cursor
		
	private var cursorSize:Number = 14;	// a square is drawn
	private var fromTop:Number = 10;	// pixels cursor is drawn from top of motion rectangle
	private var c:MovieClip; // a holder for the cursor clip
	private var cam:MovieClip; // the cam instance
	private var camW:Number; // the unscaled width of the cam
	private var camH:Number;
	private var signal:Camera; // the camera signal
	private var cm:ColorMatrixFilter; // a color matrix
	private var bf:BlurFilter; // a blur filter
	private var camCheck:Boolean = false; // check to see user accepted connection
	private var signalID; // interval id
	private var scale:Number; // cam scale taken from xscale of initial cam
	
	// ------------------ motion button vars --------------------
	
	private static var buttonNum:Number = 0; // static var for button
	
	private var b:Object;	
	private var dispatchEvent:Function;
	public var addEventListener:Function;
	public var removeEventListener:Function;	
	
	
	// ------------------ motion cursor methods --------------------
	
	function VideoMotion(theCam) {
		mx.events.EventDispatcher.initialize(this);
					
		// cam is a reference to a video instance on the stage
		// get the video instance by clicking the library menu and saying new video
		cam = theCam;	
		scale = cam._xscale / 100;
	
		// provide original cam rectangle as read properties
		left = cam._x;
		top = cam._y;
		right = left + cam._width;
		bottom = top + cam._height;
		
		// capturing bitmap always ignores any stage transformations
		// so get rid of transformations to record original size
		// then re-apply the transformations and mirror
		cam._xscale = 100;
		cam._yscale = 100;
		signal = Camera.get(0);
		cam.attachVideo(signal);			
		camW = cam._width;
		camH = cam._height;
		
		// flip the cam instance around to get a mirror effect		
		cam._xscale *= -scale;
		cam._yscale *= scale
		cam._x += cam._width;				

		// fudge to not trigger motion movement events at start
				
		_root.vm_signalID = setInterval(startstopEvents, 200);	
		function startstopEvents() {
			if (!Camera.get().muted) {				
				_root.vm_signalID2 = setInterval(startstopEvents2, 1000);	
				clearInterval(_root.vm_signalID);
			}
		}				
		function startstopEvents2() {
				_root.mv_camCheck = true;
				clearInterval(_root.vm_signalID2);
		}	
		

		// we will apply some filters so that we can better detect motion
		// we are about to set up three filters to be used for later.
		
		// first we set up a color matrix filter to up the contrast of the image
		// to do this we boost each color channel then reduce overall brightness
		// we create a color matrix that will boost each color (multiplication)
		// and then drop the brightness of the channel (addition)
		
		var boost = 4; //3
		var brightness = -50; //-60;
		var cmArray = [
			boost,0,0,0,brightness,	
			0,boost,0,0,brightness,
			0,0,boost,0,brightness,
			0,0,0,1,0
		];
		
		// create a new colorMatrixFilter so that we can apply our color matrix
		cm = new ColorMatrixFilter(cmArray);		
		
		// set up a blur filter to help emphasize areas of change
		bf = new BlurFilter(16,16,2);					
		
	}	
	
	
	public function makeCursor(theL, theT, theR, theB, theS) {	
		
		// keep track of cursors created
		cursorNum++; 		
		dispatchEvent({type: "onStart", target: this});
		
		// create a unique cursor clip that we will return
		c = _root.createEmptyMovieClip("mCursor"+cursorNum,_root.getNextHighestDepth());	
		
		if (theS == undefined) {theS = 4;}
		else if (theS < 0) {theS = 1;}
		else if (theS > 10) {theS = 10;}		
		c.trackSpeed = theS / 10;
		c.cursorSpeed = theS / 10;
		
		// create a clip that will hold the overall motion rectangle that the cursor follows
		c.motionRectangle = _root.createEmptyMovieClip("mRectangle"+cursorNum,_root.getNextHighestDepth());
		c.motionRectangle._visible = false;
		c.motionCheck = false;
		c.cursorNum = cursorNum;
		
		// draw a cursor with black outside and grey inside square	
		drawCursor(c);
	
		// set the rectangle in which the cursor will work
		c.cL = (theL != undefined) ? theL : 0; 
		c.cT = (theT != undefined) ? theT : 0;
		c.cR = (theR != undefined) ? theR : cam._width; 		  
		c.cB = (theB != undefined) ? theB : cam._height; 
		
		// drawing a camera always draws 100% so scale the rectangle down
		c.cW = (c.cR - c.cL) / scale;  c.cH = (c.cB - c.cT) / scale;		

		// we flip the camera so this takes care of that
		c.cR1 = c.cR / scale;
		c.cR2 = camW - c.cR1;
		c.cT1 = c.cT / scale;
		c.cT2 = c.cT1;		
		
		// here we figure out translations required to capture our rectangle
		c.moveX = c.cR2;
		c.moveY = c.cT1;
		
		// whenever we draw to a bitmap with a blend we need a transformation matrix
		c.m = new Matrix();					
		c.m.translate(-c.moveX, -c.moveY);
		
		// prepare two bitmap objects to store the previous and current video frames		
		c.older = new BitmapData(c.cW, c.cH, false);
		c.newer = new BitmapData(c.cW, c.cH, false);						
		
		// set some initial motion variables for the cursor
		c.motionLtotal = 0; c.motionRtotal = 0;	
		c.motionTtotal = 0;	c.motionWtotal = 0;
		c.motionTally = 0;		
		
		// an interval that will run the function follow every 100 milliseconds
		// follow puts a rectangle around motion
		c.rectID = setInterval(this, "follow", 100, c);		

		// this interval runs moveCursor which puts the cursor on the top
		// of the motion rectangle and to the sides as applicable
		c.moveID = setInterval(this, "moveCursor", c.trackSpeed * 1000, c);		
		
		return c;
	}	
	
	private function follow(c) {		
		// Generally, capture two frames across time and apply a difference filter
		// color will only show where there is movement - turn this to one color
		// draw a rectangle around the color to create a "motion rectangle"
		// Technique learned from Grant Skinner Talk at FITC 2006
	
		// We copy the picture from the old frame to a new bitmap
		c.newer.copyPixels(c.older,c.older.rectangle,null);	
		// We then draw what is currently on the camera over top of the old frame
		// As we are specifying using the difference filter, any pixels of the new
		// frame that have the same color as the old frame will have a difference of zero
		// zero means black and then every where that is not black will be some color
		c.newer.draw(cam,c.m,null,"difference"); 
		// Below we draw the unaffected camera feed to the old frame so that
		// when the follow function is called again, we have a copy of the old frame
		c.older.draw(cam,c.m,null);
		// We apply the contrast color filter to focus in on our motion
		c.newer.applyFilter(c.newer,c.newer.rectangle,null,cm);
		// We apply the blur filter to smoothen our motion region
		c.newer.applyFilter(c.newer,c.newer.rectangle,null,bf);
		// We apply a threshold to turn all colors above almost black (first number) 
		// to green (second number) the last number is a mask number (confusing)
		// this for some reason will not work unless we set the alpha channel up
		// even though we are not caring about alpha in our bitmap declaration
		// that is, the threshold would work but then the colorbounds would not
		c.newer.threshold(c.newer,c.newer.rectangle,null,">",0x00110000,0xFF00FF00,0x00FFFFFF);
		// Below we get a rectangle that encompasses the color (second number)
		// the first number is a mask (confusing because it deals with bitwise operators)
		// true means a rectangle around the color - false means a rectangle not around the color
		c.rect = c.newer.getColorBoundsRect(0x00FFFFFF,0xFF00FF00,true);
		
		// below we keep a running total of rectangle positions and a tally
		// this will allow us to average the rectangle to position the cursor
		if (c.rect.width > 0) {
			c.motionL = Math.round(cam._x - c.rect.right * scale);
			c.motionR = Math.round(cam._x - (c.rect.right - c.rect.width) * scale);
			c.motionT = Math.round(cam._y + c.rect.top * scale);
			c.motionB = Math.round(cam._y + c.rect.bottom * scale);			
			c.motionLtotal += c.motionL;
			c.motionRtotal += c.motionR;
			c.motionTtotal += c.motionT;	
			c.motionBtotal += c.motionB;	
			c.motionTally++;		
		}
	}	
	
	private function drawRect(c) {
		// used to draw the overall motion rectangle that the cursor follows
		c.mc = c.motionRectangle.createEmptyMovieClip("myRectangle", 200+c.cursorNum);
		c.mc.lineStyle(2,0xcc0000);
		c.mc.moveTo(c.mL, c.mT);
		c.mc.lineTo(c.mR, c.mT);
		c.mc.lineTo(c.mR, c.mB);
		c.mc.lineTo(c.mL, c.mB);
		c.mc.lineTo(c.mL, c.mT);
	}
	
	private function moveCursor(c) {
		// handle checking for any motion		
		if (c.motionTally > 0 && c.motionCheck == false && _root.mv_camCheck) {
			c.motionCheck = true;
			dispatchEvent({type: "onMotionStart", target: this, cursor: c});
		} else if (c.motionTally == 0 && c.motionCheck && _root.mv_camCheck) {
			c.motionCheck = false;
			dispatchEvent({type: "onMotionStop", target: this, cursor: c});				
		}
		
		// averaging cursor motion		
		c.mL = c.motionLtotal / c.motionTally - c.moveX * scale;
		c.mR = c.motionRtotal / c.motionTally - c.moveX * scale;
		c.mT = c.motionTtotal / c.motionTally + c.moveY * scale;
		c.mB = c.motionBtotal / c.motionTally + c.moveY * scale;
		
		// draw the motion rectangle - this has been set to hidden		
		drawRect(c);
		
		// get a width and a middle used in the calculation that follows
		c.mW = c.mR - c.mL;
		//c.mM = c.mR + c.mW / 2 - (cam._x - cam._width);	
		c.mM = c.mL + c.mW / 2 - (cam._x - cam._width);	
		
		//trace (cam._x - cam._width);
		
		// place cursor to left more as motion rectangle moves left
		// place cursor to right more as motion rectangle moves left
		// place cursor at the top of the motion rectangle
		c.mX = c.mL + c.mM / cam._width * c.mW;		
		c.motionLtotal = c.motionRtotal = c.motionTtotal = c.motionBtotal = c.motionTally = 0;	
		
		// tween cursor to next position
		delete(c.cursorTweenX);
		c.cursorTweenX = new Tween(c, "_x", None.easeOut, c._x, c.mX, c.cursorSpeed, true);
		delete(c.cursorTweenY);
		c.cursorTweenY = new Tween(c, "_y", None.easeOut, c._y, c.mT + fromTop, c.cursorSpeed, true);		
		
	}	
	
	private function drawCursor(c) {
		c.moveTo(-cursorSize/2,-cursorSize/2); 
		c.lineStyle(1,0x000000);		
		c.lineTo(cursorSize/2,-cursorSize/2);
		c.lineTo(cursorSize/2, cursorSize/2);
		c.lineTo(-cursorSize/2, cursorSize/2);
		c.lineTo(-cursorSize/2,-cursorSize/2); 
		c.moveTo(-cursorSize/2+1,-cursorSize/2+1); 
		c.lineStyle(1,0xCCCCCC);
		c.lineTo(cursorSize/2-1, -cursorSize/2+1);
		c.lineTo(cursorSize/2-1, cursorSize/2-1);
		c.lineTo(-cursorSize/2+1, cursorSize/2-1); 
		c.lineTo(-cursorSize/2+1,-cursorSize/2+1);			
	}
	
	
	
	
	// ------------------ motion button methods --------------------
	
	public function makeButton(cursor, clip, holdDuration) {
		
		// keep track of cursors created
		buttonNum++; 
		
		// create a new object that holds the button information
		b = _root["mButton"+buttonNum] = new Object();
		
		b.mCursor = cursor;		
		b.clip = clip;
		b.holdDuration = (holdDuration != undefined) ? holdDuration : 2;
				
		motionOver(b);
		motionOut(b);
		motionHold(b);
		
	}			
	
	// --------------  Over function ----------------------------
	
	private function motionOver(b) {		
		b.motionOverID = setInterval(this, "overFunction", 500, b);	
		b.clearCheck = true;	
	}
	private function overFunction(b) {
		if (b.mCursor.hitTest(b.clip)) {			
			if (b.clearCheck) {
				dispatchEvent({type: "onMotionOver", target: this, cursor: b.mCursor, button: b.clip});
				b.motionOnClip = b.clip;
				b.clearCheck = false;
			} 		
		} else {
			b.clearCheck = true;
		}
	}
	
	// --------------  Out function ----------------------------
	
	private function motionOut(b) {
		b.motionOutID = setInterval(this, "outFunction", 500, b);	
		b.clearCheck = true;		
	}
	private function outFunction(b) {
		if (!b.mCursor.hitTest(b.clip)) {
			if (b.motionOnClip == b.clip) {
				dispatchEvent({type: "onMotionOut", target: this, cursor: b.mCursor, button: b.clip});
				b.motionOnClip = null;
			} 		
		} 
	}	
	
	// --------------  Hold function ---------------------------- 
	
	private function motionHold(b) {
		b.motionHoldID = setInterval(this, "holdFunction", 500, b);	
		b.clearCheck = true;				
	}
	private function holdFunction(b) {
		if (b.mCursor.hitTest(b.clip)) {
			if (b.clearCheckHold) {
				clearInterval(b.holdTimeID);
				b.holdTimeID = setInterval(this, "holdTime", b.holdDuration*1000, b);	
				b.motionOnClipHold = b.clip;
				b.clearCheckHold = false;
			} 		
		} else {
			b.clearCheckHold = true;
			if (b.motionOnClipHold == b.clip) {
				clearInterval(b.holdTimeID);
				b.motionOnClipHold = null;
			} 				
		}
	}
	
	private function holdTime(b) {
		clearInterval(b.holdTimeID);
		b.motionOnClipHold = null;
		dispatchEvent({type: "onMotionHold", target: this, cursor: b.mCursor, button: b.clip});
	}	
	
	
}